@@ -1,21 +1,4 @@
 #ifndef ASMINLINE
-#endif
-/* Copyright (C) 2000  The PARI group.
-
-This file is part of the PARI/GP package.
-
-PARI/GP is free software; you can redistribute it and/or modify it under the
-terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
-ANY WARRANTY WHATSOEVER.
-
-Check the License for details. You should have received a copy of it, along
-with the package; see the file 'COPYING'. If not, write to the Free Software
-Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
-
-/*
-NOASM addll mulll bfffo divll
-*/
 #line 2 "../src/kernel/none/addll.h"
 /* Copyright (C) 2003  The PARI group.
 
@@ -31,7 +14,7 @@ with the package; see the file 'COPYING'. If not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
 /* This file originally adapted from gmp-3.1.1 (from T. Granlund), files
- * longlong.h and gmp-e_IMPL.h
+ * longlong.h and gmp-impl.h
 
   Copyright (C) 2000 Free Software Foundation, Inc. */
 
@@ -322,6 +305,128 @@ bfffo(ulong x)
 }
 #endif
 
+#endif
+#endif
+#line 2 "../src/kernel/aarch64/asm0.h"
+/* Copyright (C) 2015  The PARI group.
+
+This file is part of the PARI/GP package.
+
+PARI/GP is free software; you can redistribute it and/or modify it under the
+terms of the GNU General Public License as published by the Free Software
+Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+ANY WARRANTY WHATSOEVER.
+
+Check the License for details. You should have received a copy of it, along
+with the package; see the file 'COPYING'. If not, write to the Free Software
+Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
+
+/*
+ASM addll mulll bfffo
+NOASM divll
+*/
+
+#ifdef ASMINLINE
+#define LOCAL_HIREMAINDER  register ulong hiremainder
+#define LOCAL_OVERFLOW     register ulong overflow
+
+#define addll(a, b) \
+__extension__ ({ ulong __value, __arg1 = (a), __arg2 = (b); \
+ __asm__ ("adds %0,%2,%3\n\tadc %1,xzr,xzr\n\t" \
+   : "=&r" (__value), "=&r" (overflow) \
+   : "r" (__arg1), "r" (__arg2): "cc"); \
+ __value; \
+})
+
+#define addllx(a, b) \
+__extension__ ({ ulong __value, __arg1 = (a), __arg2 = (b); \
+ __asm__ ("subs xzr,%4,#1\n\tadcs %0,%2,%3\n\tadc %1,xzr,xzr\n\t" \
+   : "=&r" (__value), "=&r" (overflow) \
+   : "r" (__arg1), "r" (__arg2), "r" (overflow): "cc"); \
+ __value; \
+})
+
+#define addllx8(a,b,c,overflow) \
+do { long *__arg1 = a, *__arg2 = b, *__out = c; \
+     ulong __temp1, __temp2; \
+     __asm__( \
+"subs xzr,%6,#1\n\t" \
+" ldr %0, [%3]     \n\t ldr %1, [%4]    \n\t adcs %1, %0, %1\n\t str %1, [%5]    \n\t" \
+" ldr %0, [%3,-8]  \n\t ldr %1, [%4,-8] \n\t adcs %1, %0, %1\n\t str %1, [%5,-8] \n\t" \
+" ldr %0, [%3,-16] \n\t ldr %1, [%4,-16]\n\t adcs %1, %0, %1\n\t str %1, [%5,-16]\n\t" \
+" ldr %0, [%3,-24] \n\t ldr %1, [%4,-24]\n\t adcs %1, %0, %1\n\t str %1, [%5,-24]\n\t" \
+" ldr %0, [%3,-32] \n\t ldr %1, [%4,-32]\n\t adcs %1, %0, %1\n\t str %1, [%5,-32]\n\t" \
+" ldr %0, [%3,-40] \n\t ldr %1, [%4,-40]\n\t adcs %1, %0, %1\n\t str %1, [%5,-40]\n\t" \
+" ldr %0, [%3,-48] \n\t ldr %1, [%4,-48]\n\t adcs %1, %0, %1\n\t str %1, [%5,-48]\n\t" \
+" ldr %0, [%3,-56] \n\t ldr %1, [%4,-56]\n\t adcs %1, %0, %1\n\t str %1, [%5,-56]\n\t" \
+"adc %2,xzr,xzr\n\t" \
+        : "=&r" (__temp1), "=&r" (__temp2), "=&r" (overflow) \
+        : "r" (__arg1), "r" (__arg2), "r" (__out), "r" (overflow), \
+          "0" ((ulong)0), "1" ((ulong)0) \
+        : "cc"); \
+} while(0)
+
+#define subll(a, b) \
+__extension__ ({ ulong __value, __arg1 = (a), __arg2 = (b); \
+ __asm__ ("subs %0,%2,%3\n\tngc %1,xzr\n\tsub %1,xzr,%1\n\t" \
+   : "=&r" (__value), "=&r" (overflow) \
+   : "r" (__arg1), "r" (__arg2): "cc"); \
+ __value; \
+})
+
+#define subllx(a, b) \
+__extension__ ({ ulong __value, __arg1 = (a), __arg2 = (b); \
+ __asm__ ("subs xzr,xzr,%4\n\tsbcs %0,%2,%3\n\tngc %1,xzr\n\tsub %1,xzr,%1\n\t" \
+   : "=&r" (__value), "=&r" (overflow) \
+   : "r" (__arg1), "r" (__arg2), "r" (overflow): "cc"); \
+ __value; \
+})
+
+#define subllx8(a,b,c,overflow) \
+do { long *__arg1 = a, *__arg2 = b, *__out = c; \
+     ulong __temp1, __temp2; \
+     __asm__( \
+"subs xzr,xzr,%6\n\t" \
+" ldr %0, [%3]     \n\t ldr %1, [%4]    \n\t sbcs %1, %0, %1\n\t str %1, [%5]    \n\t" \
+" ldr %0, [%3,-8]  \n\t ldr %1, [%4,-8] \n\t sbcs %1, %0, %1\n\t str %1, [%5,-8] \n\t" \
+" ldr %0, [%3,-16] \n\t ldr %1, [%4,-16]\n\t sbcs %1, %0, %1\n\t str %1, [%5,-16]\n\t" \
+" ldr %0, [%3,-24] \n\t ldr %1, [%4,-24]\n\t sbcs %1, %0, %1\n\t str %1, [%5,-24]\n\t" \
+" ldr %0, [%3,-32] \n\t ldr %1, [%4,-32]\n\t sbcs %1, %0, %1\n\t str %1, [%5,-32]\n\t" \
+" ldr %0, [%3,-40] \n\t ldr %1, [%4,-40]\n\t sbcs %1, %0, %1\n\t str %1, [%5,-40]\n\t" \
+" ldr %0, [%3,-48] \n\t ldr %1, [%4,-48]\n\t sbcs %1, %0, %1\n\t str %1, [%5,-48]\n\t" \
+" ldr %0, [%3,-56] \n\t ldr %1, [%4,-56]\n\t sbcs %1, %0, %1\n\t str %1, [%5,-56]\n\t" \
+"ngc %2,xzr\n\tsub %2,xzr,%2\n\t" \
+        : "=&r" (__temp1), "=&r" (__temp2), "=&r" (overflow) \
+        : "r" (__arg1), "r" (__arg2), "r" (__out), "r" (overflow), \
+          "0" ((ulong)0), "1" ((ulong)0) \
+        : "cc"); \
+} while(0)
+
+#define mulll(a, b) \
+__extension__ ({ ulong __value, __arg1 = (a), __arg2 = (b); \
+ __asm__ ("mul %0,%1,%2\n\t" \
+   : "=r" (__value) : "r" (__arg1), "r" (__arg2)); \
+ __asm__ ("umulh %0,%1,%2\n\t" \
+   : "=r" (hiremainder) : "r" (__arg1), "r" (__arg2)); \
+ __value; \
+})
+
+#define addmul(a, b) \
+__extension__ ({ ulong __value, __arg1 = (a), __arg2 = (b), __hi; \
+ __asm__ ("madd %0,%1,%2,%3\n\t" \
+   : "=r" (__value) : "r" (__arg1), "r" (__arg2), "r" (hiremainder)); \
+ __asm__ ("umulh %0,%1,%2\n\t" \
+   : "=r" (__hi) : "r" (__arg1), "r" (__arg2)); \
+ hiremainder = (__value < hiremainder) ? __hi+1: __hi;\
+ __value; \
+})
+
+#define bfffo(a) \
+__extension__ ({ ulong __a = (a), __value; \
+    __asm__ ("clz %0, %1" : "=r" (__value) : "r" (__a)); \
+    __value; \
+})
+
 #endif
 #line 2 "../src/kernel/none/divll.h"
 /* Copyright (C) 2003  The PARI group.
@@ -338,7 +443,7 @@ with the package; see the file 'COPYING'. If not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
 /* This file originally adapted from gmp-3.1.1 (from T. Granlund), files
- * longlong.h and gmp-e_IMPL.h
+ * longlong.h and gmp-impl.h
 
   Copyright (C) 2000 Free Software Foundation, Inc. */
 
@@ -481,84 +586,275 @@ divll(ulong n0, ulong d)
 #endif /* __GNUC__ */
 
 #endif
-#ifdef LONG_IS_64BIT
-#define __MULII_KARATSUBA_LIMIT         -1
-#define __SQRI_KARATSUBA_LIMIT          -1
-#define __MULII_FFT_LIMIT               -1
-#define __SQRI_FFT_LIMIT                -1
-#define __MULRR_MULII_LIMIT             74
-#define __Fp_POW_REDC_LIMIT             17
-#define __Fp_POW_BARRETT_LIMIT         127
-#define __INVNEWTON_LIMIT              520
-#define __DIVRR_GMP_LIMIT                4
-#define __EXPNEWTON_LIMIT               66
-#define __LOGAGM_LIMIT                   6
-#define __LOGAGMCX_LIMIT                22
-#define __AGM_ATAN_LIMIT                60
-#define __INVMOD_GMP_LIMIT               3
-#define __Flx_MUL_KARATSUBA_LIMIT      142
-#define __Flx_SQR_KARATSUBA_LIMIT      316
-#define __Flx_MUL_HALFMULII_LIMIT        5
-#define __Flx_SQR_HALFSQRI_LIMIT         3
-#define __Flx_MUL_MULII_LIMIT            7
-#define __Flx_SQR_SQRI_LIMIT             5
-#define __Flx_MUL_MULII2_LIMIT           5
-#define __Flx_SQR_SQRI2_LIMIT            7
-#define __Flx_INVBARRETT_LIMIT        1142
-#define __Flx_DIVREM_BARRETT_LIMIT     768
-#define __Flx_REM_BARRETT_LIMIT       1266
-#define __Flx_BARRETT_LIMIT            398
-#define __Flx_HALFGCD_LIMIT             81
-#define __Flx_GCD_LIMIT               1017
-#define __Flx_EXTGCD_LIMIT             241
-#define __FpX_INVBARRETT_LIMIT         111
-#define __FpX_DIVREM_BARRETT_LIMIT     113
-#define __FpX_REM_BARRETT_LIMIT        111
-#define __FpX_BARRETT_LIMIT             38
-#define __FpX_HALFGCD_LIMIT             58
-#define __FpX_GCD_LIMIT                406
-#define __FpX_EXTGCD_LIMIT              87
-#define __RgX_MUL_LIMIT                  9
-#define __RgX_SQR_LIMIT                 38
+#line 2 "../src/kernel/none/divll_pre.h"
+/* Copyright (C) 2014  The PARI group.
+
+This file is part of the PARI/GP package.
+
+PARI/GP is free software; you can redistribute it and/or modify it under the
+terms of the GNU General Public License as published by the Free Software
+Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+ANY WARRANTY WHATSOEVER.
+
+Check the License for details. You should have received a copy of it, along
+with the package; see the file 'COPYING'. If not, write to the Free Software
+Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
+
+#undef  LOCAL_HIREMAINDER
+extern ulong hiremainder;
+#if defined(INLINE) && defined(__GNUC__) && !defined(DISABLE_INLINE)
+#define LOCAL_HIREMAINDER register ulong hiremainder
 #else
-#define __MULII_KARATSUBA_LIMIT         -1
-#define __SQRI_KARATSUBA_LIMIT          -1
-#define __MULII_FFT_LIMIT               -1
-#define __SQRI_FFT_LIMIT                -1
-#define __MULRR_MULII_LIMIT              8
-#define __Fp_POW_REDC_LIMIT              3
-#define __Fp_POW_BARRETT_LIMIT          11
-#define __INVNEWTON_LIMIT               66
+#define LOCAL_HIREMAINDER
+#endif
+
+#if defined(INLINE) && defined(__GNUC__) && !defined(DISABLE_INLINE)
+INLINE ulong /* precompute inverse of n */
+get_Fl_red(ulong n)
+{
+  LOCAL_HIREMAINDER;
+  n <<= bfffo(n);
+  hiremainder = ~n;
+  return divll(~0UL, n);
+}
+#else
+INLINE ulong /* precompute inverse of n */
+get_Fl_red(ulong n)
+{
+  ulong q, oldhi = hiremainder;
+  n <<= bfffo(n);
+  hiremainder = ~n;
+  q = divll(~0UL, n);
+  hiremainder = oldhi;
+  return q;
+}
+#endif
+
+INLINE ulong /* requires u1 <= n, n normalised */
+divll_pre_normalized(ulong u1, ulong u0, ulong n, ulong ninv, ulong *pt_r)
+{
+  ulong q0, q1, r;
+  LOCAL_HIREMAINDER;
+  LOCAL_OVERFLOW;
+  q0 = mulll(ninv, u1); q1 = hiremainder;
+  q0 = addll(q0, u0);
+  q1 = addllx(q1+1, u1);
+  r = u0 - q1 * n;
+  if (r > q0)
+  {
+    r += n; q1--;
+  }
+  if (r >= n)
+  {
+    r -= n; q1++;
+  }
+  *pt_r = r; return q1;
+}
+
+INLINE ulong /* requires u1 <= n, n normalised */
+remll_pre_normalized(ulong u1, ulong u0, ulong n, ulong ninv)
+{
+  ulong q0, q1, r;
+  LOCAL_HIREMAINDER;
+  LOCAL_OVERFLOW;
+  q0 = mulll(ninv, u1); q1 = hiremainder;
+  q0 = addll(q0, u0);
+  q1 = addllx(q1, u1);
+  r = u0 - (q1 + 1) * n;
+  if (r >= q0)
+    r += n;
+  return r < n ? r : r - n;
+}
+
+INLINE ulong /* reduce <a_hi, a_lo> mod n */
+remll_pre(ulong a_hi, ulong a_lo, ulong n, ulong ninv)
+{
+  int norm = bfffo(n);
+  int bits = BITS_IN_LONG - norm;
+  ulong sn = n << norm;
+  if (a_hi >= n) /* reduce a_hi first */
+  {
+    const ulong u1 = norm ? a_hi >> bits : 0;
+    const ulong u0 = a_hi << norm;
+    a_hi = remll_pre_normalized(u1, u0, sn, ninv) >> norm;
+  }
+  /* now reduce <a_hi, a_lo> */
+  {
+    const ulong u1 = ((a_hi << norm) | (norm ? a_lo >> bits: 0));
+    const ulong u0 =   a_lo << norm;
+    return remll_pre_normalized(u1, u0, sn, ninv) >> norm;
+  }
+}
+
+#if !defined(INLINE)
+extern ulong divll_pre(ulong a_lo, ulong n, ulong ninv);
+#else
+
+#if defined(__GNUC__) && !defined(DISABLE_INLINE)
+#define divll_pre(a, n, ninv)                                           \
+__extension__ ({                                                        \
+  ulong __a = (a);                                                      \
+  ulong __n = (n);                                                      \
+  int norm = bfffo(__n);                                                \
+  int bits = BITS_IN_LONG - norm;                                       \
+  ulong r, sn = __n << norm;                                            \
+  const ulong u1 = ((hiremainder << norm) | (norm ? __a >> bits: 0));   \
+  const ulong u0 = __a << norm;                                         \
+  const ulong q = divll_pre_normalized(u1, u0, sn, ninv, &r);           \
+  hiremainder = r>>norm; q;                                             \
+              })
+
+#else /* __GNUC__ */
+INLINE ulong
+divll_pre(ulong a_lo, ulong n, ulong ninv)
+{
+  int norm = bfffo(n);
+  int bits = BITS_IN_LONG - norm;
+  ulong r, sn = n << norm;
+  const ulong u1 = ((hiremainder << norm) | (norm ? a_lo >> bits: 0));
+  const ulong u0 = a_lo << norm;
+  const ulong q  = divll_pre_normalized(u1, u0, sn, ninv, &r);
+  hiremainder = r>>norm; return q;
+}
+#endif /* __GNUC__ */
+
+#endif
+#ifdef LONG_IS_64BIT
+#define __AGM_ATAN_LIMIT                 60
 #define __DIVRR_GMP_LIMIT                4
-#define __EXPNEWTON_LIMIT              197
-#define __LOGAGM_LIMIT                  45
-#define __LOGAGMCX_LIMIT                32
-#define __AGM_ATAN_LIMIT                89
-#define __INVMOD_GMP_LIMIT               3
-#define __Flx_MUL_KARATSUBA_LIMIT       90
-#define __Flx_SQR_KARATSUBA_LIMIT      159
-#define __Flx_MUL_HALFMULII_LIMIT        7
-#define __Flx_SQR_HALFSQRI_LIMIT         4
-#define __Flx_MUL_MULII_LIMIT            8
+#define __EXPNEWTON_LIMIT                66
+#define __F2x_MUL_KARATSUBA_LIMIT        15
+#define __Flx_BARRETT_HALFMULII_LIMIT    21
+#define __Flx_BARRETT_KARATSUBA_LIMIT    1172
+#define __Flx_BARRETT_MULII2_LIMIT       16
+#define __Flx_BARRETT_MULII_LIMIT        448
+#define __Flx_BARRETT_QUARTMULII_LIMIT   23
+#define __Flx_DIVREM_BARRETT_LIMIT       768
+#define __Flx_EXTGCD_LIMIT               241
+#define __Flx_GCD_LIMIT                  1017
+#define __Flx_HALFGCD_HALFMULII_LIMIT    48
+#define __Flx_HALFGCD_KARATSUBA_LIMIT    77
+#define __Flx_HALFGCD_MULII2_LIMIT       25
+#define __Flx_HALFGCD_MULII_LIMIT        71
+#define __Flx_HALFGCD_QUARTMULII_LIMIT   52
+#define __Flx_INVBARRETT_HALFMULII_LIMIT 231
+#define __Flx_INVBARRETT_KARATSUBA_LIMIT 5067
+#define __Flx_INVBARRETT_MULII2_LIMIT    26
+#define __Flx_INVBARRETT_MULII_LIMIT     1154
+#define __Flx_INVBARRETT_QUARTMULII_LIMIT 139
+#define __Flx_MUL_HALFMULII_LIMIT        5
+#define __Flx_MUL_KARATSUBA_LIMIT        142
+#define __Flx_MUL_MULII2_LIMIT           5
+#define __Flx_MUL_MULII_LIMIT            7
+#define __Flx_MUL_QUARTMULII_LIMIT       5
+#define __Flx_REM_BARRETT_LIMIT          1266
+#define __Flx_SQR_HALFSQRI_LIMIT         3
+#define __Flx_SQR_KARATSUBA_LIMIT        316
+#define __Flx_SQR_QUARTSQRI_LIMIT        3
+#define __Flx_SQR_SQRI2_LIMIT            7
 #define __Flx_SQR_SQRI_LIMIT             5
-#define __Flx_MUL_MULII2_LIMIT         152
-#define __Flx_SQR_SQRI2_LIMIT          470
-#define __Flx_INVBARRETT_LIMIT        1115
-#define __Flx_DIVREM_BARRETT_LIMIT    1289
-#define __Flx_REM_BARRETT_LIMIT        689
-#define __Flx_BARRETT_LIMIT            327
-#define __Flx_HALFGCD_LIMIT            321
-#define __Flx_GCD_LIMIT               2514
-#define __Flx_EXTGCD_LIMIT             632
-#define __FpX_INVBARRETT_LIMIT         121
-#define __FpX_DIVREM_BARRETT_LIMIT     116
-#define __FpX_REM_BARRETT_LIMIT        127
-#define __FpX_BARRETT_LIMIT             44
-#define __FpX_HALFGCD_LIMIT             55
-#define __FpX_GCD_LIMIT                414
-#define __FpX_EXTGCD_LIMIT              81
+#define __FlxqX_BARRETT_LIMIT            17
+#define __FlxqX_DIVREM_BARRETT_LIMIT     46
+#define __FlxqX_EXTGCD_LIMIT             44
+#define __FlxqX_GCD_LIMIT                470
+#define __FlxqX_HALFGCD_LIMIT            60
+#define __FlxqX_INVBARRETT_LIMIT         22
+#define __FlxqX_REM_BARRETT_LIMIT        48
+#define __FpXQX_BARRETT_LIMIT            12
+#define __FpXQX_DIVREM_BARRETT_LIMIT     30
+#define __FpXQX_EXTGCD_LIMIT             28
+#define __FpXQX_GCD_LIMIT                191
+#define __FpXQX_HALFGCD_LIMIT            35
+#define __FpXQX_INVBARRETT_LIMIT         40
+#define __FpXQX_REM_BARRETT_LIMIT        30
+#define __FpX_BARRETT_LIMIT              38
+#define __FpX_DIVREM_BARRETT_LIMIT       113
+#define __FpX_EXTGCD_LIMIT               87
+#define __FpX_GCD_LIMIT                  406
+#define __FpX_HALFGCD_LIMIT              58
+#define __FpX_INVBARRETT_LIMIT           111
+#define __FpX_REM_BARRETT_LIMIT          111
+#define __Fp_POW_BARRETT_LIMIT           127
+#define __Fp_POW_REDC_LIMIT              17
+#define __INVMOD_GMP_LIMIT               3
+#define __INVNEWTON_LIMIT                75
+#define __LOGAGMCX_LIMIT                 22
+#define __LOGAGM_LIMIT                   6
+#define __MULII_FFT_LIMIT                -1
+#define __MULII_KARATSUBA_LIMIT          -1
+#define __MULRR_MULII_LIMIT              74
+#define __RgX_MUL_LIMIT                  9
+#define __RgX_SQR_LIMIT                  38
+#define __SQRI_FFT_LIMIT                 -1
+#define __SQRI_KARATSUBA_LIMIT           -1
+#else
+#define __AGM_ATAN_LIMIT                 89
+#define __DIVRR_GMP_LIMIT                4
+#define __EXPNEWTON_LIMIT                197
+#define __F2x_MUL_KARATSUBA_LIMIT        23
+#define __Flx_BARRETT_HALFMULII_LIMIT    23
+#define __Flx_BARRETT_KARATSUBA_LIMIT    905
+#define __Flx_BARRETT_MULII2_LIMIT       647
+#define __Flx_BARRETT_MULII_LIMIT        433
+#define __Flx_BARRETT_QUARTMULII_LIMIT   20
+#define __Flx_DIVREM_BARRETT_LIMIT       1289
+#define __Flx_EXTGCD_LIMIT               632
+#define __Flx_GCD_LIMIT                  2514
+#define __Flx_HALFGCD_HALFMULII_LIMIT    78
+#define __Flx_HALFGCD_KARATSUBA_LIMIT    139
+#define __Flx_HALFGCD_MULII2_LIMIT       537
+#define __Flx_HALFGCD_MULII_LIMIT        91
+#define __Flx_HALFGCD_QUARTMULII_LIMIT   37
+#define __Flx_INVBARRETT_HALFMULII_LIMIT 240
+#define __Flx_INVBARRETT_KARATSUBA_LIMIT 3600
+#define __Flx_INVBARRETT_MULII2_LIMIT    1815
+#define __Flx_INVBARRETT_MULII_LIMIT     1293
+#define __Flx_INVBARRETT_QUARTMULII_LIMIT 73
+#define __Flx_MUL_HALFMULII_LIMIT        7
+#define __Flx_MUL_KARATSUBA_LIMIT        90
+#define __Flx_MUL_MULII2_LIMIT           152
+#define __Flx_MUL_MULII_LIMIT            8
+#define __Flx_MUL_QUARTMULII_LIMIT       7
+#define __Flx_REM_BARRETT_LIMIT          689
+#define __Flx_SQR_HALFSQRI_LIMIT         4
+#define __Flx_SQR_KARATSUBA_LIMIT        159
+#define __Flx_SQR_QUARTSQRI_LIMIT        4
+#define __Flx_SQR_SQRI2_LIMIT            470
+#define __Flx_SQR_SQRI_LIMIT             5
+#define __FlxqX_BARRETT_LIMIT            17
+#define __FlxqX_DIVREM_BARRETT_LIMIT     46
+#define __FlxqX_EXTGCD_LIMIT             44
+#define __FlxqX_GCD_LIMIT                1289
+#define __FlxqX_HALFGCD_LIMIT            89
+#define __FlxqX_INVBARRETT_LIMIT         22
+#define __FlxqX_REM_BARRETT_LIMIT        48
+#define __FpXQX_BARRETT_LIMIT            12
+#define __FpXQX_DIVREM_BARRETT_LIMIT     30
+#define __FpXQX_EXTGCD_LIMIT             28
+#define __FpXQX_GCD_LIMIT                182
+#define __FpXQX_HALFGCD_LIMIT            35
+#define __FpXQX_INVBARRETT_LIMIT         40
+#define __FpXQX_REM_BARRETT_LIMIT        30
+#define __FpX_BARRETT_LIMIT              44
+#define __FpX_DIVREM_BARRETT_LIMIT       116
+#define __FpX_EXTGCD_LIMIT               81
+#define __FpX_GCD_LIMIT                  414
+#define __FpX_HALFGCD_LIMIT              55
+#define __FpX_INVBARRETT_LIMIT           121
+#define __FpX_REM_BARRETT_LIMIT          127
+#define __Fp_POW_BARRETT_LIMIT           11
+#define __Fp_POW_REDC_LIMIT              3
+#define __INVMOD_GMP_LIMIT               3
+#define __INVNEWTON_LIMIT                93
+#define __LOGAGMCX_LIMIT                 32
+#define __LOGAGM_LIMIT                   45
+#define __MULII_FFT_LIMIT                -1
+#define __MULII_KARATSUBA_LIMIT          -1
+#define __MULRR_MULII_LIMIT              8
 #define __RgX_MUL_LIMIT                  7
-#define __RgX_SQR_LIMIT                 34
+#define __RgX_SQR_LIMIT                  34
+#define __SQRI_FFT_LIMIT                 -1
+#define __SQRI_KARATSUBA_LIMIT           -1
 #endif
 #line 2 "../src/kernel/gmp/int.h"
 /* Copyright (C) 2000  The PARI group.
@@ -643,6 +939,20 @@ evalprecp(long x)
   return v;
 }
 
+INLINE int
+varncmp(long x, long y)
+{
+  if (varpriority[x] < varpriority[y]) return  1;
+  if (varpriority[x] > varpriority[y]) return -1;
+  return 0;
+}
+INLINE long
+varnmin(long x, long y)
+{ return (varpriority[x] <= varpriority[y])? x: y; }
+INLINE long
+varnmax(long x, long y)
+{ return (varpriority[x] >= varpriority[y])? x: y; }
+
 /* Inhibit some area gerepile-wise: declare it to be a non recursive
  * type, of length l. Thus gerepile won't inspect the zone, just copy it.
  * For the following situation:
@@ -687,17 +997,18 @@ INLINE GEN
 new_chunk(size_t x) /* x is a number of longs */
 {
   GEN z = ((GEN) avma) - x;
-  if (x > (avma-bot) / sizeof(long)) pari_err(e_STACK);
   CHECK_CTRLC
+  if (x > (avma-pari_mainstack->bot) / sizeof(long))
+    new_chunk_resize(x);
   avma = (pari_sp)z;
-
 #ifdef MEMSTEP
-  if (DEBUGMEM && memused != DISABLE_MEMUSED) {
-    long d = (long)memused - (long)z;
-    if (d > 4*MEMSTEP || d < -4*MEMSTEP)
+  if (DEBUGMEM && pari_mainstack->memused != DISABLE_MEMUSED) {
+    long d = (long)pari_mainstack->memused - (long)z;
+    if (labs(d) > 4*MEMSTEP)
     {
-      memused = (pari_sp)z;
-      err_printf("...%4.0lf Mbytes used\n",(top-memused)/1048576.);
+      pari_mainstack->memused = (pari_sp)z;
+      err_printf("...%4.0lf Mbytes used\n",
+                (pari_mainstack->top-pari_mainstack->memused)/1048576.);
     }
   }
 #endif
@@ -729,7 +1040,7 @@ INLINE GEN
 cgetg_block(long x, long y)
 {
   GEN z = newblock((size_t)x);
-  z[0] = evaltyp(y) | evallg(x);
+  z[0] = CLONEBIT | evaltyp(y) | evallg(x);
   return z;
 }
 INLINE GEN
@@ -764,7 +1075,7 @@ INLINE GEN
 cgetr_block(long x)
 {
   GEN z = newblock((size_t)x);
-  z[0] = evaltyp(t_REAL) | evallg(x);
+  z[0] = CLONEBIT | evaltyp(t_REAL) | evallg(x);
   return z;
 }
 INLINE GEN
@@ -938,6 +1249,8 @@ real_0_bit(long bitprec) { GEN x=cgetr(2); x[1]=evalexpo(bitprec); return x; }
 INLINE GEN
 real_0(long prec) { return real_0_bit(-prec2nbits(prec)); }
 INLINE GEN
+real_1_bit(long bit) { return real_1(nbits2prec(bit)); }
+INLINE GEN
 real_1(long prec) {
   GEN x = cgetr(prec);
   long i;
@@ -986,7 +1299,11 @@ cmpir(GEN x, GEN y)
   GEN z;
 
   if (!signe(x)) return -signe(y);
-  if (!signe(y)) return  signe(x);
+  if (!signe(y))
+  {
+    if (expo(y) >= expi(x)) return 0;
+    return signe(x);
+  }
   av=avma; z = itor(x, realprec(y)); avma=av;
   return cmprr(z,y); /* cmprr does no memory adjustment */
 }
@@ -1006,7 +1323,7 @@ INLINE int
 cmprs(GEN x, long y) { return -cmpsr(y,x); }
 /* compare x and |y| */
 INLINE int
-cmpui(ulong x, GEN y)
+abscmpui(ulong x, GEN y)
 {
   long l = lgefint(y);
   ulong p;
@@ -1018,7 +1335,7 @@ cmpui(ulong x, GEN y)
   return p < x ? 1 : -1;
 }
 INLINE int
-cmpiu(GEN x, ulong y) { return -cmpui(y,x); }
+abscmpiu(GEN x, ulong y) { return -abscmpui(y,x); }
 INLINE int
 cmpsi(long x, GEN y)
 {
@@ -1064,13 +1381,13 @@ equalsi(long x, GEN y)
 }
 /* x == |y| ? */
 INLINE int
-equalui(ulong x, GEN y)
+absequalui(ulong x, GEN y)
 {
   if (!x) return !signe(y);
   return (lgefint(y) == 3 && (ulong)y[2] == x);
 }
 INLINE int
-equaliu(GEN x, ulong y) { return equalui(y,x); }
+absequaliu(GEN x, ulong y) { return absequalui(y,x); }
 INLINE int
 equalis(GEN x, long y) { return equalsi(y,x); }
 
@@ -1174,6 +1491,24 @@ INLINE long mod4(GEN x)  { return mod2BIL(x) & 3; }
 INLINE long mod2(GEN x)  { return mod2BIL(x) & 1; }
 INLINE int
 mpodd(GEN x) { return signe(x) && mod2(x); }
+/* x mod 2^n, n < BITS_IN_LONG */
+INLINE ulong
+umodi2n(GEN x, long n)
+{
+  long s = signe(x);
+  const ulong _2n = 1UL << n;
+  ulong m;
+  if (!s) return 0;
+  m = *int_LSW(x) & (_2n - 1);
+  if (s < 0 && m) m = _2n - m;
+  return m;
+}
+INLINE ulong Mod64(GEN x){ return umodi2n(x,6); }
+INLINE ulong Mod32(GEN x){ return umodi2n(x,5); }
+INLINE ulong Mod16(GEN x){ return umodi2n(x,4); }
+INLINE ulong Mod8(GEN x) { return umodi2n(x,3); }
+INLINE ulong Mod4(GEN x) { return umodi2n(x,2); }
+INLINE ulong Mod2(GEN x) { return umodi2n(x,1); }
 
 INLINE GEN
 truedivii(GEN a,GEN b) { return truedvmdii(a,b,NULL); }
@@ -1199,6 +1534,11 @@ smodss(long x, long y)
   long r = x%y;
   return (r >= 0)? r: labs(y) + r;
 }
+INLINE ulong
+umodsu(long x, ulong y)
+{
+  return x>=0 ? x%y: Fl_neg((-x)%y, y);
+}
 
 INLINE long
 sdivss_rem(long x, long y, long *r)
@@ -1219,6 +1559,13 @@ udivuu_rem(ulong x, ulong y, ulong *r)
   if (!y) pari_err_INV("udivuu_rem",gen_0);
   *r = x % y; return x / y;
 }
+INLINE ulong
+ceildivuu(ulong a, ulong b)
+{
+  ulong c = a/b;
+  return (a%b)? c+1: c;
+}
+
 
 INLINE ulong
 udivui_rem(ulong x, GEN y, ulong *r)
@@ -1462,6 +1809,66 @@ dvdiiz(GEN x, GEN y, GEN z)
   affii(p1,z); avma=av; return 1;
 }
 
+INLINE ulong
+remlll_pre(ulong u2, ulong u1, ulong u0, ulong n, ulong ninv)
+{
+  u1 = remll_pre(u2, u1, n, ninv);
+  return remll_pre(u1, u0, n, ninv);
+}
+
+INLINE ulong
+Fl_sqr_pre(ulong a, ulong p, ulong pi)
+{
+  register ulong x;
+  LOCAL_HIREMAINDER;
+  x = mulll(a,a);
+  return remll_pre(hiremainder, x, p, pi);
+}
+
+INLINE ulong
+Fl_mul_pre(ulong a, ulong b, ulong p, ulong pi)
+{
+  register ulong x;
+  LOCAL_HIREMAINDER;
+  x = mulll(a,b);
+  return remll_pre(hiremainder, x, p, pi);
+}
+
+INLINE ulong
+Fl_addmul_pre(ulong x0, ulong x1, ulong y0, ulong p, ulong pi)
+{
+  ulong l0, h0;
+  LOCAL_HIREMAINDER;
+  hiremainder = y0;
+  l0 = addmul(x0, x1); h0 = hiremainder;
+  return remll_pre(h0, l0, p, pi);
+}
+
+INLINE ulong
+Fl_addmulmul_pre(ulong x0, ulong y0, ulong x1, ulong y1, ulong p, ulong pi)
+{
+  ulong l0, l1, h0, h1;
+  LOCAL_OVERFLOW;
+  LOCAL_HIREMAINDER;
+  l0 = mulll(x0, y0); h0 = hiremainder;
+  l1 = mulll(x1, y1); h1 = hiremainder;
+  l0 = addll(l0, l1); h0 = addllx(h0, h1);
+  return overflow ? remlll_pre(1, h0, l0, p, pi): remll_pre(h0, l0, p, pi);
+}
+
+INLINE ulong
+Fl_ellj_pre(ulong a4, ulong a6, ulong p, ulong pi)
+{
+  /* a43 = 4 a4^3 */
+  ulong a43 = Fl_double(Fl_double(
+              Fl_mul_pre(a4, Fl_sqr_pre(a4, p, pi), p, pi), p), p);
+  /* a62 = 27 a6^2 */
+  ulong a62 = Fl_mul_pre(Fl_sqr_pre(a6, p, pi), 27 % p, p, pi);
+  ulong z1 = Fl_mul_pre(a43, 1728 % p, p, pi);
+  ulong z2 = Fl_add(a43, a62, p);
+  return Fl_div(z1, z2, p);
+}
+
 /*******************************************************************/
 /*                                                                 */
 /*                        MP (INT OR REAL)                         */
@@ -1529,6 +1936,15 @@ Fl_triple(ulong a, ulong p)
   res += a;
   return (res >= p || res < a)? res - p: res;
 }
+INLINE ulong
+Fl_halve(ulong a, ulong p)
+{
+  ulong ap, ap2;
+  if ((a&1UL)==0) return a>>1;
+  ap = a + p; ap2 = ap>>1;
+  return ap>=a ? ap2: (ap2|HIGHBIT);
+}
+
 INLINE ulong
 Fl_add(ulong a, ulong b, ulong p)
 {
@@ -1656,7 +2072,7 @@ shift_left(GEN z2, GEN z1, long imin, long imax, ulong f,  ulong sh)
     *te-- = (l << sh) | k;
     k     = l >> m;
   }
-  *te = (*se << sh) | k;
+  *te = (((ulong)*se) << sh) | k;
 }
 /* z2[imin..imax] := f.z1[imin..imax-1] shifted right sh bits
  * (feeding f from the left). Assume sh > 0 */
@@ -1737,7 +2153,7 @@ affsr(long x, GEN y)
     sh = bfffo(x);
     y[1] = evalsigne(1) | _evalexpo((BITS_IN_LONG-1)-sh);
   }
-  y[2] = x<<sh; for (i=3; i<ly; i++) y[i]=0;
+  y[2] = ((ulong)x)<<sh; for (i=3; i<ly; i++) y[i]=0;
 }
 
 INLINE void
